function [L,cpu_time,Z,eta,sample]=IRMNwayMultinomial(A,noc,opts)

% Non-parametric IRM of Bi-partite multi-graphs based on collapsed Gibbs sampling with split
% merge moves
%
% Usage: 
%  [L,cpu_time,Z,eta,sample]=IRMNwayMultinomial(A,noc,opts)
%
% Input:
%   A       n-way x types matrix of occurrences between n-domains
%   noc     1 x n  vector indicating the number of initial clusters in Z
%   opts.
%           maxiter     maximum number of iterations
%           Z           cell array of mode specific clusterings
%           dZstep      number of iteratiosn between each recorded sample
%           verbose     1: display iteration results, 0: no display
%           eta0        intial value of 1xtypes vector of pseudo link and non-link counts between the groups (default: ones(1,typess) )
%
% Output:
%   L           Log-likelihood function at each iteration ignoring constant
%               terms
%   cpu_time    cpu-time cost for each iteration
%   Z           cell array of the estimated clustering assigment matrices
%   eta         Average between group type probabilities densitites
%   sample      sampled parameters at each dZstep iteration
% 
% The method is described in:
%   Mrup, M., Gluckstad, F. K., Herlau, T., & Schmidt, M. N. (2014, September). 
%   "Nonparametric statistical structuring of knowledge systems using binary feature matches." 
%   In Machine Learning for Signal Processing (MLSP), 2014 IEEE International Workshop on (pp. 1-6). IEEE.
%
% This code is provided as is without any kind of warranty!
%
% Written by Morten Mrup
 
if nargin<3
    opts=struct;
end

N=size(A);
C=N(end);
N=N(1:end-1);
nmodes=length(N);

% Initialize Z
for k=1:nmodes
    ind=ceil(noc(k)*rand(1,N(k)));
    Z{k}=sparse(ind,1:N(k),ones(1,N(k)),noc(k),N(k));
    Z{k}=full(Z{k});
    Z{k}(sum(Z{k},2)==0,:)=[];
end

% Set remaining parameters
alpha=mgetopt(opts,'alpha',log(N));
maxiter=mgetopt(opts,'maxiter',50);
verbose=mgetopt(opts,'verbose',1);
eta0=mgetopt(opts,'eta0',ones(1,C)); % pseudo counts of links and non-links between clusters
sample_step=mgetopt(opts,'sample_step',25); % Steps between samples

% Set algorithm variables
sample=struct;
L=zeros(1,maxiter);
cpu_time=L;
sstep=0;
Q=-inf;
Qbest=-inf;
iter=0;
str2b=[];
for k=1:nmodes
    str(((k-1)*8)+1:((k-1)*8)+8)='%12.0f |';    
    str2a=['%' num2str(3*12) 's |'];
    str2b=[str2b ['noc' num2str(k)]];
end
if verbose % Display algorithm    
    disp(['Non-parametric bi-partite clustering based on the IRM model for Multinomial graphs'])
    dheader = sprintf(['%12s | %12s | %12s | ' str2a ' %12s | %12s'],'Iteration','logP','dlogP/|logP|',str2b,'time');
    dline = sprintf('-------------+--------------+--------------+--------------+--------------+--------------');
    disp(dline);
    disp(dheader);
    disp(dline);
end


% Main Loop
while iter<maxiter
   
    iter=iter+1;
    tic;
    Qold=Q;
    
    % Gibbs sampling of Z for each mode 
    for k=1:nmodes
        modes=setdiff(1:nmodes,k);
        Zm=1;
        for kk=modes
            Zm=kron(Z{kk},Zm);
        end        
        Am=reshape(permute(A,[k, modes , nmodes+1]),[N(k),prod(N(modes)),C]);            
        [Z{k},logP_A,logP_Z(k)]=Gibbs_sample_ZIRM(Z{k},Zm,Am,eta0,alpha(k),randperm(N(k)),1);            
        for t=1:10
            [Z{k},logP_A,logP_Z(k)]=split_merge_sample_Z(Z{k},Zm,Am,eta0,alpha(k),logP_A,logP_Z(k),1);
        end  
        [~,ind]=sort(sum(Z{k},2),'descend');
        Z{k}=Z{k}(ind,:);
        noc(k)=size(Z{k},1);
        
        % sample alpha        
        [logP_Z(k),alpha(k)]=sample_alpha(sum(Z{k},2),N(k),alpha(k));        
    end    
    
    % sample eta0            
    [logP_A,eta0]=sample_eta0(Z{k},Zm,Am,eta0);           
    eta=calculateEta(Zm,Z{k},Am,eta0);
    eta=permute(reshape(eta,[noc(k), noc(modes), C]),[2:nmodes 1 nmodes+1]);                
    
    % Evaluate result
    Q=logP_A+sum(logP_Z);
    dQ=Q-Qold;    
    L(iter)=Q;
    t_iter=toc;
    cpu_time(iter)=t_iter;    
   
    % Display iteration
    if rem(iter,1)==0 && verbose                
        disp(sprintf(['%12.0f | %12.4e | %12.4e | ' str '  %12.4f '],iter,Q,dQ/abs(Q),noc,t_iter));
    end         
    
    % Ztore sample
    if mod(iter,sample_step)==0         
        sstep=sstep+1;
        sample.iteration(sstep)=iter;
        sample.Z{sstep}=Z;                
        sample.eta{sstep}=eta;
        sample.alpha{sstep}=alpha;
        sample.eta0{sstep}=eta0;
    end        
    if Q>Qbest
       sample.MAP.Z=Z;        
       sample.MAP.iter=iter;
       sample.MAP.L=Q;
       sample.MAP.eta=eta;
       sample.MAP.alpha=alpha;
       sample.MAP.eta0=eta0;
       Qbest=Q;
    end
                    
end

% Display final iteration
if verbose   
  disp('Result of final iteration');
  disp(sprintf(['%12.0f | %12.4e | %12.4e | ' str '  %12.4f '],iter,Q,dQ/abs(Q),noc,t_iter));  
end

% -------------------------------------------------------------------------
% Parser for optional arguments
function var = mgetopt(opts, varname, default, varargin)
if isfield(opts, varname)
    var = getfield(opts, varname); 
else
    var = default;
end
for narg = 1:2:nargin-4
    cmd = varargin{narg};
    arg = varargin{narg+1};
    switch cmd
        case 'instrset',
            if ~any(strcmp(arg, var))
                fprintf(['Wrong argument %s = ''%s'' - ', ...
                    'Using default : %s = ''%s''\n'], ...
                    varname, var, varname, default);
                var = default;
            end
        otherwise,
            error('Wrong option: %s.', cmd);
    end
end

%-------------------------------------------------------------------------
function eta=calculateEta(Z2,Z1,A,eta0)
        noc1=size(Z1,1);
        noc2=size(Z2,1);        
        C=size(A,3);
        eta=zeros(noc1,noc2,C);                        
        ZAZt=zeros(noc1,noc2,C);   
        for c=1:C
            ZAZt(:,:,c)=Z1*A(:,:,c)*Z2'+eta0(c);                                                
        end
        eta(:,:,:)=bsxfun(@times,1./sum(ZAZt,3),ZAZt);                                                    

% -------------------------------------------------------------------------  
function [Z2,logP_A,logP_Z2]=split_merge_sample_Z(Z2,Z1,A,eta0,alpha,logP_A,logP_Z2,mode)
    %[logP_A_t,logP_Z2_t]=evalLikelihood(Z2,Z1,A,W,eta0,alpha,mode);                                 

    noc2=size(Z2,1);
    if mode==2
        J=size(A,2);    
    else
        J=size(A,1);    
    end
    
    % step 1 select two observations i and j        
    ind1=ceil(J*rand);        
    ind2=ceil((J-1)*rand);
    if ind1<=ind2
       ind2=ind2+1;
    end
    clust1=find(Z2(:,ind1));
    clust2=find(Z2(:,ind2));

    if clust1==clust2 % Split   
        setZ=find(sum(Z2([clust1 clust2],:)));    
        setZ=setdiff(setZ,[ind1,ind2]);
        n_setZ=length(setZ);
        Z2_t=Z2;
        Z2_t(clust1,:)=0;        
        comp=[clust1 noc2+1];               
        Z2_t(comp(1),ind1)=1;
        Z2_t(comp(2),ind2)=1;

        % Reassign by restricted gibbs sampling        
        if n_setZ>0
            for rep=1:3
                [Z2_t,logP_A_t,logP_Z2_t,logQ_trans,comp]=Gibbs_sample_ZIRM(Z2_t,Z1,A,eta0,alpha,setZ(randperm(n_setZ)),mode,comp);                     
            end     
        else
           logQ_trans=0;
           [logP_A_t,logP_Z2_t]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);                 
        end
        
        % Calculate Metropolis-Hastings ratio
        a_split=rand<exp(logP_A_t+logP_Z2_t-logP_A-logP_Z2-logQ_trans);  
        
        if a_split
           disp(['Splitting cluster ' num2str(clust1)])
           logP_A=logP_A_t;
           logP_Z2=logP_Z2_t;
           Z2=Z2_t;
        end
    else % Merge                                     
        Z2_t=Z2;
        Z2_t(clust1,:)=Z2_t(clust1,:)+Z2_t(clust2,:);
        setZ=find(Z2_t(clust1,:));           
        Z2_t(clust2,:)=[];        
        if clust2<clust1
            clust1_t=clust1-1;
        else 
            clust1_t=clust1;
        end
        noc2_t=noc2-1;

        % calculate likelihood of merged cluster       
        [logP_A_t,logP_Z2_t]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);                

        % Zplit the merged cluster and calculate transition probabilties                        
        setZ=setdiff(setZ,[ind1,ind2]);
        n_setZ=length(setZ);
        Z2_tt=Z2_t;
        Z2_tt(clust1_t,:)=0;        
        comp=[clust1_t noc2_t+1];               
        Z2_tt(comp(1),ind1)=1;
        Z2_tt(comp(2),ind2)=1;                
        
        % Reassign by restricted gibbs sampling
        if n_setZ>0
            for rep=1:2        
                [Z2_tt,logP_A_tt,logP_Z2_tt,logQ_trans,comp]=Gibbs_sample_ZIRM(Z2_tt,Z1,A,eta0,alpha,setZ(randperm(n_setZ)),mode,comp);                
            end
            Force=[1 2]*Z2([clust1 clust2],:);        
            [Z2_tt,logP_A_tt,logP_Z2_tt,logQ_trans]=Gibbs_sample_ZIRM(Z2_tt,Z1,A,eta0,alpha,setZ(randperm(n_setZ)),mode,comp,Force);                        
        else
            logQ_trans=0;                   
        end
        a_merge=rand<exp(logP_A_t+logP_Z2_t-logP_A-logP_Z2+logQ_trans);         
        if a_merge
          disp(['Merging cluster ' num2str(clust1) ' with cluster ' num2str(clust2)])
          logP_A=logP_A_t;
          logP_Z2=logP_Z2_t;
          Z2=Z2_t;          
        end
    end

% -------------------------------------------------------------------------  
function [logP_A,logP_Z2]=evalLikelihood(Z2,Z1,A,eta0,alpha,mode)

    my_func=@multinomialln;
    const=my_func(permute(eta0,[1 3 2]));                   
    
    if mode==2
        J=size(A,2);
    else
        J=size(A,1);
    end
    noc1=size(Z1,1);
    noc2=size(Z2,1);
    C=size(A,3);
    n_link=zeros(noc1,noc2,C);                    
    for c=1:C
        if mode==2
            n_link(:,:,c)=Z1*A(:,:,c)*Z2'+eta0(c);            
        else
            n_link(:,:,c)=Z1*A(:,:,c)'*Z2'+eta0(c);            
        end    
    end    
    sumZ2=sum(Z2,2);
    logP_A=sum(sum(my_func(n_link)))-noc1*noc2*const;    
    if nargout>1
        logP_Z2=noc2*log(alpha)+sum(gammaln(full(sumZ2)))-gammaln(J+alpha)+gammaln(alpha);      
    end
    
% -------------------------------------------------------------------------
function [Z2,logP_A,logP_Z2,logQ_trans,comp]=Gibbs_sample_ZIRM(Z2,Z1,A,eta0,alpha,JJ,mode,comp,Force)    
    
    if nargin<9
        Force=[];
    end
    if nargin<8
        comp=[];
    end    
    logQ_trans=0;

    my_func=@multinomialln; 
    const=my_func(permute(eta0,[1 3 2]));                   
        
    if mode==2
        [I,J,C]=size(A);
    else
        [J,I,C]=size(A);
    end       
    
    t=0;   
    sumZ1=sum(Z1,2);
    noc1=length(sumZ1);    
    sumZ2=sum(Z2,2);
    noc2=length(sumZ2);            
    ZA=zeros(noc1,J,C);    
    n_link=zeros(noc1,noc2,C);    
    
    
    for c=1:C
        if mode==2
            ZA(:,:,c)=Z1*A(:,:,c);                
        else
            ZA(:,:,c)=Z1*A(:,:,c)';               
        end
        ZAZt=ZA(:,:,c)*Z2';        
        n_link(:,:,c)=ZAZt+eta0(c);
    end
    
    mult_eval=my_func(n_link);    
    
    for k=JJ           
        t=t+1;
        if mod(t,5000)==0
            disp(['sampling ' num2str(t) ' out of ' num2str(J) ' nodes']);
        end
        
        % Remove effect of Z2(:,k)        
        sumZ2=sumZ2-Z2(:,k);   
        Z1AZ2k=ZA(:,k,:);        
        d=find(Z2(:,k));        
        if ~isempty(d)
            n_link(:,d,:)=n_link(:,d,:)-Z1AZ2k;                                                        
            Z2(:,k)=0;               
        end
        
        if isempty(comp)
            if sumZ2(d)==0 % Remove singleton cluster            
                v=1:noc2;
                v(d)=[];
                d=[];
                noc2=noc2-1;                               
                P=sparse(1:noc2,v,ones(1,noc2),noc2,noc2+1);                        
                Z2=P*Z2;                        
                sumZ2=sumZ2(v,1);            
                n_link=n_link(:,v,:);                                       
                mult_eval=mult_eval(:,v);            
            end                                       

            % Calculate probability for existing communties as well as proposal cluster                                                                 
            mult_eval(:,d)=my_func(n_link(:,d,:)); % removed the constant -my_func(Ap,An)))                                           
            sum_mult_eval=sum(sum(mult_eval));
            e=ones(1,noc2);
            sum_mult_eval_d=sum_mult_eval-[sum(mult_eval,1) 0];
            n_link_new=zeros(size(Z1AZ2k,1),1,C);
            for c=1:C
                n_link_new(:,1,c)=Z1AZ2k(:,1,c)+eta0(c);
            end
            TTT1=cat(2,n_link+Z1AZ2k(:,e,:),n_link_new);
            
            mult_eval_d=my_func(TTT1);                                        
            logQ=sum_mult_eval_d'+[sum(mult_eval_d(:,1:noc2),1) (sum(mult_eval_d(:,noc2+1,:),1)-noc1*const)]'; % removed the constant -my_func(Ap,An)))                                             

            % Zample from posterior                        
            QQ=exp(logQ-max(logQ));
            weight=[sumZ2; alpha];
%             % code test
%             noc1t=ceil(size(Z2,1)*rand);
%             noc2t=ceil((size(Z2,1)+1)*rand);
%             Z2_t=Z2;
%             Z2_t(noc1t,k)=1;
%             [logP_A_t1,logP_Z2_t1]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);     
%             Z2_t=Z2;
%             Z2_t(noc2t,k)=1;
%             [logP_A_t2,logP_Z2_t2]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);                 
%             a1=logP_A_t1+logP_Z2_t1-(logP_A_t2+logP_Z2_t2)                        
%             a2=logQ(noc1t)+log(weight(noc1t))-(logQ(noc2t)+log(weight(noc2t)))
%             difference=a1-a2

            QQ=weight.*QQ;        
            ind=find(rand<cumsum(QQ/sum(QQ)),1,'first');     
            Z2(ind,k)=1;   
            if ind>noc2            
                noc2=noc2+1;
                sumZ2(noc2,1)=0;
                for c=1:C
                    n_link(:,noc2,c)=eta0(c);                                                
                end
                mult_eval(:,noc2)=0;                      
            end
            mult_eval_d=mult_eval_d(:,ind);            
        else            
            % Calculate probability for existing communties as well as proposal cluster                                                                  
            mult_eval(:,d)=my_func(n_link(:,d,:)); % removed the constant -my_func(Ap,An)))                                           
            sum_mult_eval=sum(sum(mult_eval));                        
            e=ones(1,2);
            sum_mult_eval_d=sum_mult_eval-sum(mult_eval(:,comp,:),1);
            mult_eval_d=my_func(n_link(:,comp,:)+Z1AZ2k(:,e,:));                                        
            logQ=sum_mult_eval_d'+sum(mult_eval_d,1)'; % removed the constant -my_func(Ap,An)))                                             
           
            % Zample from posterior                        
            QQ=exp(logQ-max(logQ));
            weight=sumZ2(comp);
            
%                         % code test
%             noc1t=comp(1);
%             noc2t=comp(2);
%             Z2_t=Z2;
%             Z2_t(noc1t,k)=1;
%             [logP_A_t1,logP_Z2_t1]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);     
%             Z2_t=Z2;
%             Z2_t(noc2t,k)=1;
%             [logP_A_t2,logP_Z2_t2]=evalLikelihood(Z2_t,Z1,A,eta0,alpha,mode);                 
%             a1=logP_A_t1+logP_Z2_t1-(logP_A_t2+logP_Z2_t2)                        
%             a2=logQ(1)+log(weight(1))-(logQ(2)+log(weight(2)))
%             difference=a1-a2

            
            
            QQ=weight.*QQ;
            QQ=QQ/sum(QQ);            
            if isempty(Force)
                ind=find(rand<cumsum(QQ),1,'first');
            else 
                ind=Force(k);
            end
            q_tmp=logQ-max(logQ)+log(weight);
            q_tmp=q_tmp-log(sum(exp(q_tmp)));            
            logQ_trans=logQ_trans+q_tmp(ind);
            Z2(comp(ind),k)=1;  
            mult_eval_d=mult_eval_d(:,ind);            
            ind=comp(ind);
        end
                        
        % Re-enter effect of new s_k        
        sumZ2=sumZ2+Z2(:,k);
        n_link(:,ind,:)=n_link(:,ind,:)+Z1AZ2k;                                            
        mult_eval(:,ind)=mult_eval_d;
                
        % Remove empty clusters        
        if ~all(sumZ2)
            d=find(sumZ2==0);
            if ~isempty(comp)
                ind_d=find(d<comp);
                comp(ind_d)=comp(ind_d)-1;
            end
            v=1:noc2;
            v(d)=[];
            noc2=noc2-length(d);                               
            P=sparse(1:noc2,v,ones(1,noc2),noc2,noc2+1);                        
            Z2=P*Z2;                        
            sumZ2=sumZ2(v,1);            
            n_link=n_link(:,v,:);            
            mult_eval=mult_eval(:,v);            
        end              
    end      
    % Calculate Likelihood for sampled solution     
    logP_Z2=noc2*log(alpha)+sum(gammaln(full(sumZ2)))-gammaln(J+alpha)+gammaln(alpha);
    logP_A=sum(sum(mult_eval))-noc1*noc2*const;     
    
% ---------------------
function C=multinomialln(A)
        C = sum(gammaln(A),3)-gammaln(sum(A,3));        
        
    %--------------------------------------------------------------------
function [logZ,alpha]=sample_alpha(sumZ,N,alpha)

max_iter=100;
K=length(sumZ);
const=sum(gammaln(sumZ));
logZ=K*log(alpha)+const-gammaln(N+alpha)+gammaln(alpha);
accept=0;
for sample_iter=1:max_iter  
    alpha_new=exp(log(alpha)+0.1*randn); % symmetric proposal distribution in log-domain (use change of variable in acceptance rate alpha_new/alpha) 
    logZ_new=K*log(alpha_new)+const-gammaln(N+alpha_new)+gammaln(alpha_new); 
    if rand<(alpha_new/alpha*exp(logZ_new-logZ))
        alpha=alpha_new;
        logZ=logZ_new;
        accept=accept+1;       
    end
 end
 %disp(['accepted ' num2str(accept) ' out of ' num2str(max_iter) ' samples for alpha']);
    
 
 %--------------------------------------------------------------------
function [logP,eta0]=sample_eta0(Z,Zm,Am,eta0)
       
max_iter=10;
noc1=size(Z,1);
[noc2, J]=size(Zm);
C=length(eta0);
ZA=zeros(noc1,J,C);    
n_link_noeta0=zeros(noc1,noc2,C);           
for c=1:C
    ZA(:,:,c)=Z*Am(:,:,c);                   
    n_link_noeta0(:,:,c)=ZA(:,:,c)*Zm';            
end

n_link=bsxfun(@plus,n_link_noeta0,permute(eta0,[1 3 2]));
const=multinomialln(permute(eta0,[1 3 2]));                   
logP=sum(sum(multinomialln(n_link)))-noc1*noc2*const;     
accept=0;
for k=1:length(eta0)    
    for sample_iter=1:max_iter           
        eta_new=exp(log(eta0(k))+0.1*randn); % symmetric proposal distribution in log-domain (use change of variable in acceptance rate alpha_new/alpha)         
        eta0_new=eta0;
        eta0_new(k)=eta_new;
        const_new=multinomialln(permute(eta0_new,[1 3 2]));                   
        n_link_new=n_link;
        n_link_new(:,:,k)=n_link_noeta0(:,:,k)+eta_new;
        logP_new=sum(sum(multinomialln(n_link_new)))-noc1*noc2*const_new;     
        
      %  % test code
      %  [logP_A_tmp,logP_Z2]=evalLikelihood(Zm,Z,Am,eta0,1,2);
      %  [logP_A_new_tmp,logP_Z2]=evalLikelihood(Zm,Z,Am,eta0_new,1,2);
      %  logP_new-logP
      %  logP_A_new_tmp-logP_A_tmp
        
        if rand<(eta_new/eta0(k)*exp(logP_new-logP))
            eta0(k)=eta_new;
            logP=logP_new;
            n_link=n_link_new;
            accept=accept+1;       
        end
    end
 end
 %disp(['accepted ' num2str(accept) ' out of ' num2str(max_iter*length(eta0)) ' samples for eta0']);
    